const ApiConnector = require('./api_connector');
const Config = require('./config');
const { Datapoint } = require('./compiled').proto;
const CLS = require('continuation-local-storage');

/**
 * Entry point for starting the BitSensor Web Application Security plugin.
 * @class BitSensor
 */
class BitSensor {
  /**
   * Creates an instance of BitSensor.
   * @memberof BitSensor
   */
  constructor() {
    this.preHandlers = [];
    this.postHandlers = [];

    this.session = CLS.createNamespace('bitsensor');
  }

  /**
   * Runs a session with callback
   * @param {any} callback 
   * @memberof BitSensor
   */
  runSession(callback) {
    this.session.run(() => {
      if (!this.isInitialized()) {
        this.initialize();
      }
      this.traceIncrement();
      if (typeof callback === 'function') {
        callback();
      }
      this.ship();
    });
  }

  /**
   * Initializes BitSensor
   * @memberof BitSensor
   */
  initialize() {
    if (this.isInitialized()) {
      return;
    }

    this.setDatapoint(new Datapoint());
    this.initializeEnabled();
    this.runPreHandlers();
  }

  /**
   * Check if BitSensor is initialized
   * @returns {boolean}
   * @memberof BitSensor
   */
  isInitialized() {
    return this.session && !!this.session.get('datapoint');
  }

  /**
   * 
   * @param {any} consumer 
   * @memberof BitSensor
   */
  apply(consumer) {
    if (this.hasBeenDisabled()) {
      return;
    }

    if (!this.isInitialized()) {
      this.runSession(() => {
        consumer.call(this, this.getDatapoint());
      });
      return;
    }

    consumer.call(this, this.getDatapoint());
  }

  /**
   * Get the datapoint
   * @returns {Datapoint}
   * @memberof BitSensor
   */
  getDatapoint() {
    return bitSensor.session.get('datapoint');
  }

  /**
   * Set the datapoint
   * @param {Datapoint} datapoint
   * @memberof BitSensor
   */
  setDatapoint(datapoint) {
    this.session.set('datapoint', datapoint);
  }

  /**
   * Ship data to BitSensor
   * @memberof BitSensor
   */
  ship() {
    if (this.hasBeenDisabled()) {
      return;
    }

    if (this.traceDecrement() > 0) {
      return;
    }

    this.runPostHandlers();
    this.log();
    this.clear();
  }

  /**
   * Log data to BitSensor
   * @memberof BitSensor
   */
  log() {
    if (!this.config) {
      return;
    }
    new ApiConnector().from(this.config.user, this.config.apiKey)
      .to(this.config.uri)
      .with(this.getDatapoint())
      .post(ApiConnector.ACTION_LOG)
      .send();
  }

  /**
   * Clean BitSensor
   * @memberof BitSensor
   */
  clear() {
    this.session.set('isDisabled', null);
    this.session.set('isHandlingRequest', null);
    this.session.set('datapoint', null);
  }

  /**
   * Check if BitSensor has been disabled
   * @returns {boolean}
   * @memberof BitSensor
   */
  hasBeenDisabled() {
    if (this.getDisabled() == null) {
      return false;
    }

    return this.getDisabled() > 0;
  }

  /**
   * Disable BitSensor
   * @returns {boolean}
   * @memberof BitSensor
   */
  setDisabled() {
    this.initializeEnabled();
    this.session.set('isDisabled', this.getDisabled() + 1);

    return this.getDisabled();
  }

  /**
   * Enable BitSensor
   * @returns {boolean}
   * @memberof BitSensor
   */
  setEnabled() {
    this.initializeEnabled();
    if (this.getDisabled() > 0) {
      this.session.set('isDisabled', this.getDisabled() - 1);
    }

    return this.getDisabled();
  }

  /**
   * Initialize enabled BitSensor
   * @memberof BitSensor
   */
  initializeEnabled() {
    if (this.getDisabled() == null) {
      this.session.set('isDisabled', 0);
    }
    if (this.getHandlingRequest() == null) {
      this.session.set('isHandlingRequest', 0);
    }
  }

  /**
   * Check if BitSensor is disabled
   * @returns {boolean}
   * @memberof BitSensor
   */
  getDisabled() {
    return bitSensor.session.get('isDisabled');
  }

  /**
   * Check if BitSensor is handling requests
   * @returns {boolean}
   * @memberof BitSensor
   */
  getHandlingRequest() {
    return bitSensor.session.get('isHandlingRequest');
  }

  /**
   * Increments the request counter
   * @memberof BitSensor
   */
  traceIncrement() {
    if (this.getHandlingRequest() == null) {
      this.initialize();
    }

    bitSensor.session.set('isHandlingRequest', bitSensor.getHandlingRequest() + 1);
  }

  /**
   * Decrements the request counter
   * @memberof BitSensor
   */
  traceDecrement() {
    if (this.getHandlingRequest() == null) {
      this.initialize();
    }
    if (bitSensor.getHandlingRequest() > 0) {
      bitSensor.session.set('isHandlingRequest', bitSensor.getHandlingRequest() - 1);
    }

    return this.getHandlingRequest();
  }

  /**
   * Add a prehandler
   * @param {any} handler 
   * @memberof BitSensor
   */
  addPreHandler(handler) {
    this.preHandlers.push(handler);
  }

  /**
   * Add a posthandler
   * @param {any} handler 
   * @memberof BitSensor
   */
  addPostHandler(handler) {
    this.postHandlers.push(handler);
  }

  /**
   * Execute prehandlers
   * @memberof BitSensor
   */
  runPreHandlers() {
    this.preHandlers.forEach((handler) => {
      handler.handle(this.getDatapoint());
    });
  }

  /**
   * Execute posthandlers
   * @memberof BitSensor
   */
  runPostHandlers() {
    this.postHandlers.forEach((handler) => {
      handler.handle(this.getDatapoint());
    });
  }

  /**
   * Set the config
   * @param {any} config 
   * @memberof BitSensor
   */
  setConfig(config) {
    this.config = new Config(config);
  }
}

module.exports = BitSensor;
